<?php
declare(strict_types=1);
/**
 * @package qfq
 * @author kputyr
 * @date: 19.04.2024
 */

namespace IMATHUZH\Qfq\Core\Typo3;

require_once(__DIR__ . '/../../../vendor/autoload.php');

use IMATHUZH\Qfq\Core\Store\Session;
use Psr\Http\Message\ServerRequestInterface;
use Psr\Log\LoggerAwareInterface;
use Psr\Log\LoggerAwareTrait;
use TYPO3\CMS\Core\Authentication\AbstractAuthenticationService;
use TYPO3\CMS\Core\Configuration\ExtensionConfiguration;
use TYPO3\CMS\Core\Crypto\Random;
use TYPO3\CMS\Core\Database\Connection;
use TYPO3\CMS\Core\Database\ConnectionPool;
use TYPO3\CMS\Core\Database\Query\Restriction\HiddenRestriction;
use TYPO3\CMS\Core\Http\ServerRequestFactory;
use TYPO3\CMS\Core\Utility\GeneralUtility;

/**
 * This class allows to authenticate user by Qfq and synchronize it with Typo3.
 *
 * In order to authenticate a user by Qfq, create a sip variable feUser with
 * the username and request a page with parameter logintype=login. It a local
 * user with this username exists, it will be logged in.
 */
class QfqAuthenticationService extends AbstractAuthenticationService implements LoggerAwareInterface
{
    use LoggerAwareTrait;

    const AUTH_USER_SIP = 'authuser';
    const AUTH_CONF_SIP = 'authconf';

    protected ?array $user = null;

    protected ?Connection $connection = null;

    protected array $configuration;

    public function __construct(ExtensionConfiguration $extensionConfiguration)
    {
        $this->configuration = $extensionConfiguration->get('qfq');
    }

    /**
     * Finds a user provided by QFQ to authenticate. Returns `null` if none is found
     * and an array of user properties otherwise, with a boolean 'qfq/auth' property
     * to indicate whether the user should be authenticated or rejected.
     *
     * @return array|null
     * @throws \CodeException
     * @throws \Doctrine\DBAL\DBALException
     * @throws \Doctrine\DBAL\Driver\Exception
     */
    public function getUser(): ?array
    {
        $query = $this->getRequest()->getQueryParams();
        $userSip = $query[self::AUTH_USER_SIP] ?? null;
        $confSip = $query[self::AUTH_CONF_SIP] ?? null;

        // This user is not managed by QFQ
        if (!$userSip) {
            return null;
        }

        // Read user properties and service configuration passed in the sip store
        $this->openSipStore();
        $this->user = $this->validateUser($this->loadSipContent($userSip));
        if (!$this->user) {
            return null;
        }
        $config = array_merge(
            $this->getDefaultConfig(),
            $confSip ? $this->loadSipContent($confSip) : []
        );

        // Find the user or create one otherwise if allowed to do so
        $this->connect();
        $t3user = $this->findTypo3User(['username' => $this->user['username']], $config['storage']);
        if (empty($t3user) && $config['create']) {
            $this->logger->info("no local user - creating one");
            $t3user = $this->insertTypo3User($this->user, $config);
        }

        // Check if the user is disabled and react accordingly
        if ($t3user) {
            if ($t3user['disable'] && $config['unlock']) {
                $this->logger->info("the user is disabled - unlocking");
                $this->unlockTypo3User($t3user['uid']);
                $t3user['disable'] = 0;
            }
            $t3user['qfq/auth'] = !$t3user['disable'];
        } else {
            $t3user = ['qfq/auth' => false];
        }
        return $t3user;
    }

    /**
     * Authenticates the user if it managed by this service.
     *
     * @param array $user
     * @return int
     */
    public function authUser(array $user): int
    {
        if (!isset($user['qfq/auth'])) {
            return 100;     // another service should authenticate the user
        } elseif ($user['qfq/auth']) {
            return 200;     // the user is authenticated
        } else {
            return 0;       // the user is locked or cannot be created
        }
    }

    /**
     * The default configuration of the service taken from the settings of the extension.
     *
     * @return array
     */
    protected function getDefaultConfig(): array
    {
        return [
            'create' => (bool)$this->configuration['authCreateUser'],
            'unlock' => (bool)$this->configuration['authUnlockUser'],
            'storage' => $this->configuration['authUserStorage'],
            'groups' => $this->configuration['authUserGroups']
        ];
    }

    /**
     * Replaces the user with `null` if it does not contain a username.
     *
     * @param array|null $user
     * @return array|null
     */
    protected function validateUser(?array $user): ?array
    {
        return empty($user) || empty($user['username']) ? null : $user;
    }

    /**
     * The current request
     *
     * @return ServerRequestInterface
     */
    protected function getRequest(): ServerRequestInterface
    {
        return $GLOBALS['TYPO3_REQUEST'] ?? ServerRequestFactory::fromGlobals();
    }

    /**
     * Prepares a connection to the database.
     * @return void
     */
    protected function connect(): void
    {
        $this->connection = GeneralUtility::makeInstance(ConnectionPool::class)
            ->getConnectionForTable($this->pObj->user_table);
    }

    /**
     * Searches for a user in fe_users matching all the criteria from $condition.
     * If $pages is provided, it must be the comma-separated list of uid of pages
     * on which the user must be located. Otherwise, the page is not checked.
     *
     * @param array $condition   an associated array of pairs (field => value)
     * @param string $pages      a comma-separated list of uids
     * @return array|null
     * @throws \Doctrine\DBAL\DBALException
     * @throws \Doctrine\DBAL\Driver\Exception
     */
    protected function findTypo3User(array $condition, string $pages = ''): ?array
    {
        $query = $this->connection->createQueryBuilder();
        $query->getRestrictions()->removeByType(HiddenRestriction::class);
        $expr = $query->expr();
        $conditions = array_map(
            fn ($field, $value) => $expr->eq($field, $query->quote($value)),
            array_keys($condition),
            array_values($condition)
        );
        if ($pages) {
            $conditions[] = $expr->in('pid', $query->quote($pages));
        }
        return $query->select('*')
                    ->from($this->pObj->user_table)
                    ->where(...$conditions)
                    ->setMaxResults(1)
                    ->executeQuery()
                    ->fetchAssociative() ?: null;
    }

    /**
     * Inserts a new user to the fe_users tables.
     *
     * This method filters out silently properties with no matching table column.
     * No checks are made whether required properties are present except the following:
     * - `pid` is takes from $config['storage']
     * - `groups` is takes from $config['groups']
     * - `password` is assigned a random value unless already provided
     * - `crdate` and `tstamp` are the current timestamps
     *
     * @param array $properties   properties of the user
     * @param array $config
     * @return array
     */
    protected function insertTypo3User(array $properties, array $config): array
    {
        $fields = array_intersect_key(
            $properties,
            $GLOBALS['TCA'][$this->pObj->user_table]['columns']
        );
        // intval() evaluates the first number in a comma-separated list
        // ignoring the rest or returns 0 if the list is empty
        $fields['pid'] = intval($config['storage']);
        $fields['crdate'] = $GLOBALS['EXEC_TIME'];
        $fields['tstamp'] = $GLOBALS['EXEC_TIME'];
        $fields[$this->pObj->usergroup_column] = $config['groups'];
        if (!isset($fields['password'])) {
            $fields['password'] = GeneralUtility::makeInstance(Random::class)
                ->generateRandomHexString(32);
        }
        $this->connection->insert($this->pObj->user_table, $fields);
        return $this->findTypo3User(['uid' => $this->connection->lastInsertId()]);
    }

    /**
     * Unlocks a disabled user with a provided uid.
     *
     * @param int $uid
     * @return void
     */
    protected function unlockTypo3User(int $uid): void
    {
        $this->connection->update(
            $this->pObj->user_table,
            ['disable' => 0],
            ['uid' => $uid]
        );
    }

    /**
     * Initiates the session in order to access sip values
     * @return void
     * @throws \CodeException
     */
    protected function openSipStore(): void
    {
        // This initiates the PHP session
        Session::getInstance();
    }

    /**
     * Returns the content of a given sip omitting empty values.
     *
     * @param string $sip
     * @return array
     */
    protected function loadSipContent(string $sip): array
    {
        parse_str($_SESSION['qfq'][$sip] ?? '', $result);
        return array_filter($result, fn($value) => $value != '');
    }
}
